-- Widgets created by Adirelle

local AceGUI = LibStub("AceGUI-3.0")

-- Localization
local L_ACTIONTYPE = {
    spell = "Spell",
    macro = "Macro",
    item = "Item",
    equipmentset = "Equipment Set",
    battlepet = "Battle Pet",
    flyout = "Flyout Menu",
    mount = "Mount",
}

local widgetVersion = 2

--------------------------------------------------------------------------------
-- Flyout Helper
--------------------------------------------------------------------------------
local function FindFlyoutSlot(flyoutId)
    local iii = 1
    while true do
        local sbtype, sbid = GetSpellBookItemInfo(iii, "spell")
        if sbtype == "FLYOUT" and sbid == flyoutId then
            return iii
        elseif sbtype == nil and sbid == nil then
            return nil
        end
        iii = iii + 1
    end
end

local function FindFlyoutInfo(flyoutId)
    local flyoutSlot = FindFlyoutSlot(flyoutId)
    if flyoutSlot then
        return GetSpellBookItemName(flyoutSlot, "spell"), GetSpellBookItemTexture(flyoutSlot, "spell")
    else
        return nil, nil
    end
end

--------------------------------------------------------------------------------
-- Abtract action slot
--------------------------------------------------------------------------------

local BaseConstructor

do

    local function OnAcquire(self)
        self:SetWidth(200)
        self:SetHeight(44)
        self:SetDisabled(false)
        self:SetLabel()
    end

    local function OnRelease(self)
        self.rejectIcon:Hide()
        self.frame:ClearAllPoints()
        self.frame:Hide()
        self:SetDisabled(false)
        self:SetText()
    end

    local function Pickup(actionType, actionData)
        if actionType == "item" then
            return PickupItem(actionData)
        elseif actionType == "macro" then
            return PickupMacro(actionData)
        elseif actionType == "equipmentset" then
            return PickupEquipmentSetByName(actionData)
        elseif actionType == "battlepet" then
            return C_PetJournal.PickupPet(actionData)
        elseif actionType == "mount" then
            if tonumber(actionData) == 0 or tonumber(actionData) == LBK.SUMMON_RANDOM_FAVORITE_MOUNT_ID then
                return C_MountJournal.Pickup(0)
            end
            return C_MountJournal.Pickup(LBK:FindMountIndexByMountId(actionData))
        elseif actionType == "spell" then
            return PickupSpell(actionData)
        elseif actionType == "flyout" then
            return PickupSpellBookItem(FindFlyoutSlot(actionData), "spell")
        end
        ClearCursor()
    end

    LBK.Pickup = Pickup;

    local function Link(actionType, actionData)
        if actionType == "item" then
            return select(2, GetItemInfo(actionData))
        elseif actionType == "spell" then
            return GetSpellLink(actionData)
        else
            return nil
        end
    end

    local function ParseActionInfo(actionType, data1, data2, data3)
        if actionType == "mount" then
            if (data1 == LBK.SUMMON_RANDOM_FAVORITE_MOUNT_ID) then
                data1 = 0
            end
            return true, "mount", data1
        elseif actionType == "companion" then
            local _, spellId = C_MountJournal.GetDisplayedMountInfo(data1)
            return true, "spell", spellId
        elseif actionType == "battlepet" then
            return true, "battlepet", data1
        elseif actionType == "spell" then
            local link = GetSpellLink(data3)
            if link then
                return true, "spell", data3
            end
        elseif actionType == "item" then
            return true, "item", data1
        elseif actionType == "macro" then
            return true, "macro", (GetMacroInfo(data1))
        elseif actionType == "equipmentset" then
            return true, "equipmentset", data1
        elseif actionType == "flyout" then
            return true, "flyout", data1
        elseif actionType == "action" then
            return ParseActionInfo(GetActionInfo(data1))
        end
        return actionType and true or false
    end

    local function ParseCursorInfo()
        return ParseActionInfo(GetCursorInfo())
    end

    local function Button_OnEnter(this)
        local self = this.obj

        if not this.disabled then
            local hasAction, actionType = ParseCursorInfo()
            if hasAction and not self:AcceptActionType(actionType) then
                self.rejectIcon:Show()
            else
                self.rejectIcon:Hide()
            end
        end

        self:Fire("OnEnter")

        local link = Link(self.actionType, self.actionData)
        if link then
            GameTooltip:SetOwner(this, "ANCHOR_RIGHT")
            GameTooltip:SetHyperlink(link)
        elseif self.actionType == "mount" then
            GameTooltip:SetOwner(this, "ANCHOR_RIGHT")
            if tonumber(self.actionData) == 0 or tonumber(self.actionData) == LBK.SUMMON_RANDOM_FAVORITE_MOUNT_ID then
                GameTooltip:SetMountBySpellID(LBK.SUMMON_RANDOM_FAVORITE_MOUNT_SPELL);
            else
                local _, spellId = C_MountJournal.GetMountInfoByID(tonumber(self.actionData))
                GameTooltip:SetMountBySpellID(spellId)
            end
        elseif self.actionType == "flyout" then
            GameTooltip:SetOwner(this, "ANCHOR_RIGHT")
            GameTooltip:SetSpellBookItem(FindFlyoutSlot(self.actionData), "spell")
        end
    end

    local function Button_OnLeave(this)
        local self = this.obj
        self.rejectIcon:Hide()
        self:Fire("OnLeave")

        if (GameTooltip:IsOwned(this)) then
            GameTooltip:Hide()
        end
    end

    local function SetNewAction(self, newType, newData)
        local oldType, oldData = self.actionType, self.actionData
        if newType ~= oldType or newData ~= oldData then
            local value = newType and newData and self:BuildValue(newType, newData)
            self:Fire("OnEnterPressed", value)
            if self.actionType ~= oldType or self.actionData ~= oldData then
                Pickup(oldType, oldData)
            end
        end
    end

    local function Button_OnReceiveDrag(this)
        if this.disabled then return end
        local self = this.obj
        local hasAction, actionType, actionData = ParseCursorInfo()
        if hasAction and actionType and actionData and self:AcceptActionType(actionType) then
            SetNewAction(self, actionType, actionData)
        end
    end

    local function Button_OnDragStart(this)
        if this.disabled then
            local actionType, actionData = this.obj.actionType, this.obj.actionData
            if nil ~= actionType and nil ~= actionData then
                Pickup(actionType, actionData)
            end
            return
        end
        SetNewAction(this.obj)
    end

    local function Button_OnClick(this, button)
        if this.disabled then return end
        if button == "RightButton" then
            this.obj:Fire("OnEnterPressed", "")
        else
            return Button_OnReceiveDrag(this)
        end
    end

    local function SetText(self, text)
        local actionType, actionData, name, texture
        if text and text ~= "" then
            actionType, actionData, name, texture = self:ParseValue(tostring(text))
        end
        if actionType and actionData and name and texture and self:AcceptActionType(actionType) then
            self.actionType, self.actionData = actionType, actionData
            self.button:SetNormalTexture([[Interface\Buttons\UI-Quickslot2]])
            self.icon:SetTexture(texture)
            self:SetActionText(L_ACTIONTYPE[actionType] or actionType, name)
            self.text:Show()
            self.icon:Show()
        else
            self.actionType, self.actionData = nil, nil
            self.button:SetNormalTexture([[Interface\Buttons\UI-Quickslot]])
            self.text:Hide()
            self.icon:Hide()
        end
    end

    local function OnHeightSet(self, height)
        local button = self.button
        local icon = self.icon
        local tex = button:GetNormalTexture()
        local size = height - 8
        if self.label:IsShown() then
            size = size - self.label:GetHeight()
        end
        size = math.min(size, 36)
        button:SetWidth(size)
        button:SetHeight(size)
        tex:SetWidth(size * 60 / 36)
        tex:SetHeight(size * 60 / 36)
        icon:SetWidth(size)
        icon:SetHeight(size)
    end

    local function SetLabel(self, label)
        if label and label ~= "" then
            self.label:SetText(label)
            self.label:Show()
        else
            self.label:SetText("")
            self.label:Hide()
        end
        OnHeightSet(self, self.frame:GetHeight())
    end

    local function SetDisabled(self, disabled)
        self.button.disabled = disabled
        if disabled then
            self.icon:SetDesaturated(true)
            self.label:SetTextColor(0.5, 0.5, 0.5)
            self.text:SetTextColor(0.5, 0.5, 0.5)
        else
            self.icon:SetDesaturated(false)
            self.label:SetTextColor(1, .82, 0)
            self.text:SetTextColor(1, 1, 1)
        end
    end

    function BaseConstructor(self)
        self.OnRelease = OnRelease
        self.OnAcquire = OnAcquire

        self.OnWidthSet = OnWidthSet
        self.OnHeightSet = OnHeightSet

        self.SetLabel = SetLabel
        self.SetText = SetText
        self.SetDisabled = SetDisabled

        local frame = CreateFrame("Frame")
        frame:SetWidth(200)
        frame:SetHeight(44)
        frame.obj = self
        self.frame = frame

        local label = frame:CreateFontString(nil, "OVERLAY", "GameFontNormal")
        label:SetPoint("TOPLEFT")
        label:SetPoint("TOPRIGHT")
        label:SetJustifyH("LEFT")
        label:SetHeight(15)
        self.label = label

        local button = CreateFrame("Button", nil, frame)
        button:RegisterForClicks("LeftButtonUp", "RightButtonUp")
        button:RegisterForDrag("LeftButton")
        button:SetPoint("BOTTOMLEFT", 4, 4)
        button:SetWidth(29)
        button:SetHeight(29)
        button:SetNormalTexture([[Interface\Buttons\UI-Quickslot]])
        button:SetPushedTexture([[Interface\Buttons\UI-Quickslot-Depress]])
        button:GetPushedTexture():SetBlendMode('ADD')
        button:SetHighlightTexture([[Interface\Buttons\ButtonHilight-Square]])
        button:GetHighlightTexture():SetBlendMode('ADD')
        button:SetScript('OnDragStart', Button_OnDragStart)
        button:SetScript('OnReceiveDrag', Button_OnReceiveDrag)
        button:SetScript('OnClick', Button_OnClick)
        button:SetScript('OnEnter', Button_OnEnter)
        button:SetScript('OnLeave', Button_OnLeave)
        button.obj = self
        self.button = button

        local tex = button:GetNormalTexture()
        tex:ClearAllPoints()
        tex:SetPoint("CENTER")

        local icon = button:CreateTexture("BACKGROUND")
        icon:SetPoint("CENTER")
        self.icon = icon

        local rejectIcon = frame:CreateTexture("OVERLAY")
        rejectIcon:SetTexture([[Interface\PaperDollInfoFrame\UI-GearManager-LeaveItem-Transparent]])
        rejectIcon:SetAllPoints(icon)
        rejectIcon:Hide()
        self.rejectIcon = rejectIcon

        local text = frame:CreateFontString(nil, "OVERLAY", "GameFontHighlightSmall")
        text:SetPoint("TOPLEFT", button, "TOPRIGHT", 2, 0)
        text:SetPoint("BOTTOMRIGHT")
        text:SetJustifyH("LEFT")
        text:SetJustifyV("TOP")
        --text:SetWordWrap(false)
        self.text = text

        self:OnHeightSet(44)

        AceGUI:RegisterAsWidget(self)
        return self
    end
end

--------------------------------------------------------------------------------
-- Generic action slot
--------------------------------------------------------------------------------

do
    local function AcceptActionType(self, actionType)
        return actionType == "item" or actionType == "spell" or actionType == "macro" or actionType == "equipmentset" or actionType == "battlepet" or actionType == "mount" or actionType == "flyout"
    end

    local function BuildValue(self, actionType, actionData)
        return strjoin(":", tostringall(actionType, actionData))
    end

    local function ParseValue(self, value)
        local itemId = tonumber(value:match("item:(%d+)"))
        if itemId then
            local name, _, _, _, _, _, _, _, _, texture = GetItemInfo(itemId)
            return "item", itemId, name, texture
        end
        local spellId = tonumber(value:match("spell:(%d+)"))
        if spellId then
            local name, _, texture = GetSpellInfo(spellId)
            return "spell", spellId, name, texture
        end
        local flyoutId = tonumber(value:match("flyout:(%d+)"))
        if flyoutId then
            local name, texture = FindFlyoutInfo(flyoutId)
            return "flyout", flyoutId, name, texture
        end
        local macroName = value:match("macro:([^:]+)")
        if macroName then
            local name, texture = GetMacroInfo(macroName:trim())
            return "macro", name, name, texture
        end
        local battlePetGuid = value:match("battlepet:([^:]+)") or value:match("summonpet:([^:]+)")
        if battlePetGuid then
            local _, customName, _, _, _, _, _, petName, texture = C_PetJournal.GetPetInfoByPetID(battlePetGuid:trim())
            return "battlepet", battlePetGuid, (customName or petName), texture
        end
        local mountId = value:match("companion:([^:]+)")
        if mountId then
            -- rewrite the value to be a mount action with a mount id
            value = "mount:" .. LBK:FindMountIdBySpellId(mountId)
        end
        local mountId = value:match("summonmount:([^:]+)") or value:match("mount:([^:]+)")
        if mountId then
            if tonumber(mountId) == 0 or tonumber(mountId) == LBK.SUMMON_RANDOM_FAVORITE_MOUNT_ID then
                local name, _, texture = GetSpellInfo(LBK.SUMMON_RANDOM_FAVORITE_MOUNT_SPELL)
                return "mount", mountId, name, texture
            end
            local name, _, texture = C_MountJournal.GetMountInfoByID(mountId:trim())
            return "mount", mountId, name, texture
        end
        local setName = value:match("equipmentset:([^:]+)")
        if setName then
            setName = setName:trim()
            for i = 1, GetNumEquipmentSets() do
                local name, texture = GetEquipmentSetInfo(i)
                if setName == name then
                    return "equipmentset", name, name, texture
                end
            end
        end
    end

    LBK.ParseValue = ParseValue;

    local function SetActionText(self, actionType, name)
        self.text:SetFormattedText("%s: %s", actionType, name)
    end

    local widgetType = "ActionSlot"

    local function Constructor()
        return BaseConstructor {
            type = widgetType,
            AcceptActionType = AcceptActionType,
            ParseValue = ParseValue,
            BuildValue = BuildValue,
            SetActionText = SetActionText,
        }
    end

    AceGUI:RegisterWidgetType(widgetType, Constructor, widgetVersion)
end
